#!/usr/bin/python2.5

"""Robotic Chanter Controller

Copyright 2009 Neil Fraser
http://neil.fraser.name/hardware/chanter/

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

  http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
"""

import sys
import lpt
import time

sys.path.insert(0, "midi")
from MidiOutStream import MidiOutStream
from MidiInFile import MidiInFile
del sys.path[0]

PORT = lpt.lpt1

NOTE_LOW_G =  (False, False, False, False, False, False, False, False)
NOTE_LOW_A =  (False, False, False, False, False, False, False, True)
NOTE_B =      (False, False, False, False, False, False, True,  True)
NOTE_C =      (False, False, False, False, False, True,  True,  False)
NOTE_D =      (False, False, False, False, True,  True,  True,  False)
NOTE_E =      (False, False, False, True,  False, False, False, True)
NOTE_F =      (False, False, True,  True,  False, False, False, True)
NOTE_HIGH_G = (False, True,  True,  True,  False, False, False, True)
NOTE_HIGH_A = (True,  True,  True,  False, False, False, False, True)

GRACE_LOW_G =  (False, False, False, False, False, False, False, False)
GRACE_LOW_A =  (False, False, False, False, False, False, False, True)
GRACE_B =      (False, False, False, False, False, False, True,  None)
GRACE_C =      (False, False, False, False, False, True,  None,  None)
GRACE_D =      (False, False, False, False, True,  None,  None,  None)
GRACE_E =      (False, False, False, True,  None,  None,  None,  None)
GRACE_F =      (False, False, True,  None,  None,  None,  None,  None)
GRACE_HIGH_G = (False, True,  None,  None,  None,  None,  None,  None)
GRACE_HIGH_A = (True,  None,  None,  None,  None,  None,  None,  None)

NOTES = (NOTE_LOW_G, NOTE_LOW_A, NOTE_B, NOTE_C, NOTE_D, NOTE_E, NOTE_F, NOTE_HIGH_G, NOTE_HIGH_A)
GRACES = (GRACE_LOW_G, GRACE_LOW_A, GRACE_B, GRACE_C, GRACE_D, GRACE_E, GRACE_F, GRACE_HIGH_G, GRACE_HIGH_A)

NOTE_MATRIX = {}
GRACE_MATRIX = {}

PC_SPEAKER = True

Duration = 0

if PC_SPEAKER:
  import winsound
# Frequencies of: G3, A3, B3, C#4, D4, E4, F#4, G4, A4.
FREQS = (196, 220, 247, 277, 294, 330, 370, 392, 440)
FREQ_MATRIX = {}

class MidiPlayer(MidiOutStream):
  def note_on(self, channel=0, note=0x40, velocity=0x40):
    if channel == 0:
      global Duration
      Duration = self.rel_time()

  def note_off(self, channel=0, note=0x40, velocity=0x40):
    if channel == 0:
      # rel_time is normally 0 for note_off, except for the very last note.
      global Duration
      Duration += self.rel_time()
      if Duration > 100:
        matrix = NOTE_MATRIX
      else:
        matrix = GRACE_MATRIX
        # Artistic Decison: Warp the grace notes to be slightly longer.
        Duration += 20
      if matrix.has_key(note):
        setNote(matrix[note])
        print " %d %dms \t%s" % (note, Duration, "\t".join([str(x) for x in matrix[note]]))
      else:
        print "Unknown note: %d" % note
      if PC_SPEAKER:
        winsound.Beep(FREQ_MATRIX[note], Duration)
      else:
        time.sleep(Duration / 1000.0)


class MidiScanner(MidiOutStream):
  def note_on(self, channel=0, note=0x40, velocity=0x40):
    if channel == 0:
      global NOTE_MATRIX
      NOTE_MATRIX[note] = True

def setNote(note):
  # Interleaved connector
  (b7, b6, b5, b4, b3, b2, b1, b0) = note
  note = (b3, b7, b2, b6, b1, b5, b0, b4)

  byte = lpt.getByte(PORT)
  for b in xrange(8):
    mask = 2 ** b
    if note[b] == False:
      byte = byte & (mask ^ 255)
    elif note[b] == True:
      byte = byte | mask
  lpt.setByte(PORT, byte)

def main():
  if len(sys.argv) != 2:
    print >> sys.stderr, "Usage:  %s <FILENAME>" % sys.argv[0]
    sys.exit(2)
  filename = sys.argv[1]
  resetPort()
  time.sleep(1)
  try:
    if scanMidi(filename):
      playMidi(filename)
  finally:
    resetPort()

def scanMidi(filename):
  global NOTE_MATRIX, GRACE_MATRIX, FREQ_MATRIX
  NOTE_MATRIX = {}
  GRACE_MATRIX = {}
  FREQ_MATRIX = {}
  event_handler = MidiScanner()
  midi_in = MidiInFile(event_handler, filename)
  midi_in.read()
  noteValues = NOTE_MATRIX.keys()
  noteValues.sort()
  print "Found %d note values: %s" % (len(noteValues), noteValues)
  if len(noteValues) != len(NOTES):
    # We are missing some notes.  Try to match against a known scale.
    scale1 = (55, 57, 59, 61, 62, 64, 66, 67, 69)
    scale2 = (56, 58, 60, 62, 63, 65, 67, 68, 70)
    if isSubset(scale1, noteValues):
      noteValues = scale1
    elif isSubset(scale2, noteValues):
      noteValues = scale2
    else:
      print "Need %d note values.  Can't play this tune." % len(NOTES)
      return False
  for x in xrange(len(NOTES)):
    NOTE_MATRIX[noteValues[x]] = NOTES[x]
    GRACE_MATRIX[noteValues[x]] = GRACES[x]
    FREQ_MATRIX[noteValues[x]] = FREQS[x]
  return True

def isSubset(set, subset):
  # Are all the elements of subset in set?
  for x in subset:
    if not x in set:
      return False
  return True

def playMidi(filename):
  event_handler = MidiPlayer()
  midi_in = MidiInFile(event_handler, filename)
  midi_in.read()

def resetPort():
  # Set both the data port and the command port to 0.
  lpt.setByte(PORT, 0)
  lpt.setByte(PORT + 2, 0)
  
def fingerTest(pause):
  resetPort()
  time.sleep(pause)
  note = [False, False, False, False, False, False, False, False]
  for d in xrange(8):
    note[7 - d] = True
    setNote(note)
    print "%d: %s" % (d, note)
    note[7 - d] = False
    time.sleep(pause)
  resetPort()

if __name__ == "__main__":
  main()
  #fingerTest(5)
